# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

# Built-in
import os
import logging

# PyPI
import click

# Freemelt
from .. import FileHandler
from . import segment_handler, svg2obp

log = logging.getLogger(__name__)


@click.group()
def OBPC():
    pass


@OBPC.command()
@click.argument("filename")
def compile(**kwargs):
    f0 = kwargs["filename"]
    f1 = f0[:-5] + ".obp"
    log.info("Compiling {0} into {1}...".format(f0, f1))
    blob = FileHandler.get_segment_from_dict(f0)
    if os.path.isfile(f1):
        log.warning(
            "A file named {0} already exists. Would you like to overwrite it? [Y/N]".format(
                f1
            )
        )
        choice = input().lower()
        if choice == "y":
            FileHandler.write_obp(blob, f1)
        else:
            pass
    elif not os.path.isfile(f1):
        FileHandler.write_obp(blob, f1)


@OBPC.command()
@click.argument("filename")
def decompile(**kwargs):
    f0 = kwargs["filename"]
    f1 = f0[:-4] + ".obpj"
    print("Decompiling {0} into {1}...".format(f0, f1))
    blob = FileHandler.read_obp(f0)
    if os.path.isfile(f1):
        print(
            "A file named {0} already exists. Would you like to overwrite it? [Y/N]".format(
                f1
            )
        )
        choice = input().lower()
        if choice == "y":
            FileHandler.write_obpj(blob, f1)
        else:
            pass
    elif not os.path.isfile(f1):
        FileHandler.write_obpj(blob, f1)


@OBPC.command()
@click.argument("filename")
def convert(**kwargs):
    f0 = kwargs["filename"]
    f1 = f0[:-4] + ".obpj"
    print("Converting {0} into {1}...".format(f0, f1))
    paths, w, h = svg2obp.read_svg(f0)
    l = svg2obp.convert_svg(paths, w, h)
    FileHandler.write_obpj(l, f1)


@OBPC.command()
@click.argument("filename")
def validate(**kwargs):
    l = FileHandler.read_obp(kwargs["filename"])
    for seg in l:
        segment_handler.validate_segment(seg)


if __name__ == "__main__":
    OBPC()



# SPDX-FileCopyrightText: 2022 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""OBP data viewer.

usage: obpviewer.py [-h] [--slice-size SLICE_SIZE] [--index INDEX] obp_file

Use the mouse or keyboard to navigate.

The main frame vizualises the obp objects. The `--slice-size` decides
how big chunk of objects to plot. The `--index` decides the "latest"
obp object to plot. The latest position of the latest obp object is
highlighted with a white star marker. The colorbar shows the speed
used when the objects are drawn.

supported objects:
  Line, AcceleratingLine, Curve, AcceleratingCurve,
  Restore and SyncPoint.

partially supported:
  TimedPoints (NOTE: the timedpoints are visualized as tiny diamonds (visible
                     when you zoom). This is only a limitation of the
                     visualization - the points are truly points. Another
                     limitation is that the timedpoint's duration/time
                     component is not vizualised. They will be colored as
                     "0 m/s" according to the colorbar.)

not supported:
  Metadata and vendor_setup. These objects are ignored.

keyboard shortcuts:

  right        step forward
  shift+right  step forward 10 steps
  ctrl+right   step forward 100 steps
  alt+right    step forward 1000 steps

  left         step backward
  shift+left   step backward 10 steps
  ctrl+left    step backward 100 steps
  alt+left     step backward 1000 steps

  p            same as right
  n            same as left

  a            jump to start of file
  e            jump to end of file

  s            jump to spot size change
  b            jump to beam power change
  r            jump to restore point

  0-9          jump to sync point change (digit specifies
               sync point: 1st, 2nd, ..., etc)

positional arguments:
  obp_file              Path to obp file.

optional arguments:
  -h, --help            show this help message and exit
  --slice-size SLICE_SIZE
                        Initial slice size (default: 100).
  --index INDEX         Initial index (default: 100).
"""

# Built-in
import argparse
import dataclasses
import pathlib
import gzip
import sys
import tkinter
from tkinter import ttk

# Freemelt
from obplib import OBP_pb2 as obp

# PyPI
try:
    import matplotlib
except ModuleNotFoundError as error:
    sys.exit(
        "Error: matplotlib is not installed. Try:\n"
        "  $ sudo apt install python3-matplotlib\n"
        "or\n"
        "  $ python3 -m pip install matplotlib"
    )

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (
    FigureCanvasTkAgg,
    NavigationToolbar2Tk,
)
from matplotlib.figure import Figure
from matplotlib.ticker import EngFormatter
from matplotlib.path import Path
import matplotlib.collections as mcoll

import numpy as np
from google.protobuf.internal.decoder import _DecodeVarint32

plt.style.use("dark_background")


def load_obp_objects(filepath):
    """Deserialize obp data and yield protobuf messages"""
    with open(filepath, "rb") as fh:
        data = fh.read()
    if filepath.suffix == ".gz":
        data = gzip.decompress(data)
    consumed = new_pos = 0
    while consumed < len(data):
        msg_len, new_pos = _DecodeVarint32(data, consumed)
        msg_buf = data[new_pos : new_pos + msg_len]
        consumed = new_pos + msg_len
        packet = obp.Packet()
        packet.ParseFromString(msg_buf)
        attr = packet.WhichOneof("payload")
        yield getattr(packet, attr)


@dataclasses.dataclass
class Data:
    paths: list
    speeds: list
    spotsizes: list
    beampowers: list
    syncpoints: dict
    restores: list


class TimedPoint:
    pass


def _unpack_tp(obp_objects):
    # Treat each point in TimedPoints as a separate obp object.  This
    # is a hack to make TimedPoints fit in the existing logic without
    # too much redesign.
    for obj in obp_objects:
        if isinstance(obj, obp.TimedPoints):
            t = 0
            for point in obj.points:
                tp = TimedPoint()
                tp.x = point.x
                tp.y = point.y
                if point.t == 0:
                    point.t = t
                tp.t = t = point.t
                tp.params = obj.params
                yield tp
        else:
            yield obj


def load_artist_data(obp_objects) -> Data:
    """Return data used when drawing matplotlib artists"""
    paths = list()
    speeds = list()
    spotsizes = list()
    beampowers = list()
    syncpoints = dict()
    _lastseen = dict()  # last seen sync points
    restores = list()
    _restore = 0
    for obj in _unpack_tp(obp_objects):
        if isinstance(obj, (obp.Line, obp.AcceleratingLine)):
            paths.append(
                Path(
                    np.array([[obj.x0, obj.y0], [obj.x1, obj.y1]]) / 1e6,
                    (Path.MOVETO, Path.LINETO),
                )
            )
        elif isinstance(obj, TimedPoint):
            paths.append(
                # Draw a diamond
                Path(
                    np.array(
                        [
                            [obj.x - 100, obj.y],
                            [obj.x, obj.y + 100],
                            [obj.x + 100, obj.y],
                            [obj.x, obj.y - 100],
                            [obj.x - 100, obj.y],
                            [obj.x, obj.y],
                        ]
                    )
                    / 1e6,
                    (
                        Path.MOVETO,
                        Path.LINETO,
                        Path.LINETO,
                        Path.LINETO,
                        Path.LINETO,
                        Path.MOVETO,
                    ),
                )
            )
        elif isinstance(obj, (obp.Curve, obp.AcceleratingCurve)):
            paths.append(
                Path(
                    np.array(
                        [
                            [obj.p0.x, obj.p0.y],
                            [obj.p1.x, obj.p1.y],
                            [obj.p2.x, obj.p2.y],
                            [obj.p3.x, obj.p3.y],
                        ]
                    )
                    / 1e6,
                    [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4],
                )
            )
        elif isinstance(obj, obp.SyncPoint):
            if obj.endpoint not in syncpoints:
                # Catch up with zeros when seen for the first time
                syncpoints[obj.endpoint] = [0] * len(paths)
            # Update _lastseen with endpoint's new sync value
            _lastseen[obj.endpoint] = int(obj.value)
            continue
        elif isinstance(obj, obp.Restore):
            _restore = 1
            continue
        else:
            continue
        if isinstance(obj, (obp.Line, obp.Curve)):
            speeds.append(obj.speed / 1e6)
        elif isinstance(obj, (obp.AcceleratingLine, obp.AcceleratingCurve)):
            speeds.append(obj.sf)
        elif isinstance(obj, TimedPoint):
            speeds.append(0)
        else:
            speeds.append(0)
        spotsizes.append(obj.params.spot_size)
        beampowers.append(obj.params.beam_power)
        for k, v in _lastseen.items():
            syncpoints[k].append(v)
        restores.append(_restore)
        _restore = 0
    for key in syncpoints:
        syncpoints[key] = np.array(syncpoints[key])
    if len(paths) == 0:
        raise Exception("no lines or curves in obp data")
    return Data(
        paths,
        np.array(speeds),
        np.array(spotsizes),
        np.array(beampowers),
        syncpoints,
        np.array(restores),
    )


class ObpFrame(ttk.Frame):
    def __init__(self, master, data, slice_size, index=None, **kwargs):
        """Initialization of ObpFrame"""
        super().__init__(master, **kwargs)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.data = data
        if index is None:
            index = slice_size

        def cap(i):  # Prevents IndexError
            return max(0, min(len(self.data.paths) - 1, int(i)))

        self.cap = cap

        index = cap(index)
        slice_ = slice(cap(index + 1 - slice_size), cap(index) + 1)

        # Matplotlib artists and canvas
        fig = Figure(figsize=(9, 8), constrained_layout=True)
        ax = fig.add_subplot(111)
        ax.axhline(0, linewidth=1, zorder=0)  # horizontal center line
        ax.axvline(0, linewidth=1, zorder=0)  # vertical center line
        radius = 0.05  # meters
        ax.set_xlim([-radius, radius])
        ax.set_ylim([-radius, radius])
        si_meter = EngFormatter(unit="m")
        ax.xaxis.set_major_formatter(si_meter)
        ax.yaxis.set_major_formatter(si_meter)
        ax.tick_params(axis="x", labelsize=8)
        ax.tick_params(axis="y", labelsize=8)

        self.path_collection = mcoll.PathCollection(
            self.data.paths[slice_],
            facecolors="none",
            transform=ax.transData,
            cmap=plt.cm.rainbow,
            norm=plt.Normalize(vmin=0, vmax=max(self.data.speeds)),
        )
        self.path_collection.set_array(self.data.speeds[slice_])
        ax.add_collection(self.path_collection)

        ticks = list(sorted(set(self.data.speeds)))
        if len(ticks) > 10:
            ticks = None
        cbar = fig.colorbar(
            self.path_collection,
            ax=ax,
            pad=0,
            aspect=60,
            format=EngFormatter(unit="m/s"),
            ticks=ticks,
        )
        cbar.ax.tick_params(axis="x", labelsize=8)
        cbar.ax.tick_params(axis="y", labelsize=8)
        seg = self.data.paths[index]
        self.marker = ax.scatter(*seg.vertices[-1], c="white", marker="*", zorder=2)

        self.canvas = canvas = FigureCanvasTkAgg(fig, master=self)
        canvas.draw()
        canvas.mpl_connect("key_press_event", self.keypress)

        # Slice size
        self._slice_size = tkinter.IntVar(value=slice_size)
        self._slice_size_spinbox = ttk.Spinbox(
            self,
            from_=0,
            to=len(self.data.paths) - 1,
            textvariable=self._slice_size,
            command=self.update_index,
            width=6,
        )
        self._slice_size_spinbox.bind("<KeyRelease>", self.update_index)

        # Index
        self._index = tkinter.IntVar(value=index)
        self._index_scale = tkinter.Scale(
            self,
            from_=0,
            to=len(self.data.paths) - 1,
            orient=tkinter.HORIZONTAL,
            variable=self._index,
            command=self.update_index,
        )
        self._index_spinbox = ttk.Spinbox(
            self,
            from_=0,
            to=len(self.data.paths) - 1,
            textvariable=self._index,
            command=self.update_index,
            width=6,
        )
        self._index_spinbox.bind("<KeyRelease>", self.update_index)

        self.toolbar_frame = ttk.Frame(master=self)
        toolbar = NavigationToolbar2Tk(canvas, self.toolbar_frame)
        toolbar.update()

        self.info_value = tkinter.StringVar(value=",  ".join(self.get_info(index)))
        self.info_label = ttk.Label(self, textvariable=self.info_value)

        self.button_quit = tkinter.Button(self, text="Quit", command=self.master.quit)

    def keypress(self, event):
        stepsize = 1
        parts = event.key.split("+")
        if len(parts) == 2:
            prefix, key = parts
        else:
            prefix, key = "", parts[0]

        if prefix == "shift" or key in {"P", "N"}:
            stepsize = 10
        elif prefix == "ctrl":
            stepsize = 100
        elif prefix == "alt":
            stepsize = 1000
        if key in {"right", "p", "P"}:
            self._index.set(self.cap(self._index.get() + stepsize))
        elif key in {"left", "n", "N"}:
            self._index.set(self.cap(self._index.get() - stepsize))
        elif event.key == "a":
            self._index.set(0)
        elif event.key == "e":
            self._index.set(len(self.data.paths))
        elif event.key.isdigit():
            n = int(event.key)
            for i, key in enumerate(self.data.syncpoints):
                if i + 1 == n:
                    self.nextdifferent(self.data.syncpoints[key])
        elif event.key == "r":
            self.nextdifferent(self.data.restores)
        elif event.key == "b":
            self.nextdifferent(self.data.beampowers)
        elif event.key == "s":
            self.nextdifferent(self.data.spotsizes)
        else:
            print(event.key)
            return
        self.update_index()

    def nextdifferent(self, array):
        start = self.cap(self._index.get())
        bools = array[start:] != array[start]
        self._index.set(start + np.argmax(bools))

    def update_index(self, new_index=None):
        if new_index is None:
            new_index = self._index.get()
        elif isinstance(new_index, tkinter.Event):
            if new_index.keysym != "Return":
                return
            self.canvas.get_tk_widget().focus_force()
            new_index = self._index.get()
        index = self.cap(new_index)
        ss = self._slice_size.get() or 1

        # Update artists
        slice_ = slice(self.cap(index + 1 - ss), self.cap(index) + 1)
        segs = self.data.paths[slice_]
        if len(segs):
            self.path_collection.set_paths(segs)
            self.path_collection.set_array(self.data.speeds[slice_])
            self.marker.set_offsets(segs[-1].vertices[-1])
            self.canvas.draw()

        # Update labels
        self.info_value.set(",  ".join(self.get_info(index)))

    def get_info(self, index):
        info = [f"{k}={v[index]}" for k, v in self.data.syncpoints.items()]
        info.append(f"Restore={int(self.data.restores[index])}")
        info.append(f"BeamPower={int(self.data.beampowers[index])}")
        info.append(f"SpotSize={int(self.data.spotsizes[index])}")
        return info

    def setup_grid(self):
        self.canvas.get_tk_widget().grid(row=0, columnspan=4, sticky="NSWE")

        self._index_scale.grid(row=1, columnspan=4, sticky="NSWE")

        self.info_label.grid(row=2, column=0, sticky="SW")
        self._slice_size_spinbox.grid(row=2, column=1, sticky="SE")
        self._index_spinbox.grid(row=2, column=2, sticky="SE")
        self.button_quit.grid(row=2, column=3, sticky="SE")

        self.toolbar_frame.grid(row=3, columnspan=4, sticky="NSWE")


parser = argparse.ArgumentParser(
    # Copy this description text to the module docstring when modified
    description="""OBP data viewer.

Use the mouse or keyboard to navigate.

The main frame vizualises the obp objects. The `--slice-size` decides
how big chunk of objects to plot. The `--index` decides the "latest"
obp object to plot. The latest position of the latest obp object is
highlighted with a white star marker. The colorbar shows the speed
used when the objects are drawn.

supported objects:
  Line, AcceleratingLine, Curve, AcceleratingCurve,
  Restore and SyncPoint.

partially supported:
  TimedPoints (NOTE: the timedpoints are visualized as tiny diamonds (visible
                     when you zoom). This is only a limitation of the
                     visualization - the points are truly points. Another
                     limitation is that the timedpoint's duration/time
                     component is not vizualised. They will be colored as
                     "0 m/s" according to the colorbar.)

not supported:
  Metadata and vendor_setup. These objects are ignored.

keyboard shortcuts:

  right        step forward
  shift+right  step forward 10 steps
  ctrl+right   step forward 100 steps
  alt+right    step forward 1000 steps

  left         step backward
  shift+left   step backward 10 steps
  ctrl+left    step backward 100 steps
  alt+left     step backward 1000 steps

  p            same as right
  n            same as left

  a            jump to start of file
  e            jump to end of file

  s            jump to spot size change
  b            jump to beam power change
  r            jump to restore point

  0-9          jump to sync point change (digit specifies
               sync point: 1st, 2nd, ..., etc)
""",
    formatter_class=argparse.RawTextHelpFormatter,
)
parser.add_argument(
    "obp_file",
    type=pathlib.Path,
    help="Path to obp file.",
)
parser.add_argument(
    "--slice-size",
    type=int,
    default=100,
    help="Initial slice size (default: %(default)s).",
)
parser.add_argument(
    "--index",
    type=int,
    default=100,
    help="Initial index (default: %(default)s).",
)


def _main(args):

    obp_objects = load_obp_objects(args.obp_file)
    data = load_artist_data(obp_objects)

    root = tkinter.Tk()
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    root.title(f"obpviewer - {args.obp_file.name}")
    root.option_add("*tearOff", tkinter.FALSE)
    frame = ObpFrame(root, data, args.slice_size, args.index)
    frame.grid(row=0, column=0, sticky="NSWE", padx=5, pady=5)
    frame.setup_grid()
    tkinter.mainloop()


def main():
    args = parser.parse_args()
    _main(args)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""@PathSegment docstring
An abstract base class containing all methods that are common among the different path segments. Line, Curve, etc.
"""
# PyPI
from google.protobuf.message import Message


class PathSegment:
    def __init__(self):
        pass

    def get_segment_length(self) -> float:
        pass

    def get_segment_duration(self) -> float:
        pass

    def get_pb(self) -> Message:
        pass

    def get_packet(self) -> bytes:
        pass

    def write_obp(self) -> bytes:
        pass

    def get_obpj(self) -> dict:
        pass



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


"""@Point docstring

A class representing a point on a 2D grid with coordinates (x,y)
"""


# Built-in
import math


class Point:
    def __init__(self, x, y):
        """
        Contructor for the point class

        Args:
            x (Int): The x-coordinate for the point
            y (Int): The y-coordinate for the point
        """
        self.x = x
        self.y = y

    def get_x(self):
        """
        Coordinate getter method

        Returns:
            x (Int): the x-coordinate of the specific point instance
        """
        return self.x

    def get_y(self):
        """
        Coordinate getter method

        Returns:
            y (Int): the y-coordinate of the specific point instance
        """
        return self.y

    def translate(self, V):
        """
        Returns a new point that has been translated according to the coordinates in the given vector V

        Args:
            V (Vector): A translation vector

        Returns:
            Point: A point that has been translated according to V
        """
        x1 = self.get_x() + V.get_x()
        y1 = self.get_y() + V.get_y()
        return Point(x1, y1)

    def rotate(self, theta):
        """
        Returns a new Point that has been rotated theta radians relative to the origin

        Args:
            theta (Int): Number of radians to rotate

        Returns:
            Point: A point that has been rotated theta degrees around the origin
        """
        x1 = self.get_x() * math.cos(theta) - self.get_y() * math.sin(theta)
        y1 = self.get_x() * math.sin(theta) + self.get_y() * math.cos(theta)
        return Point(x1, y1)

    def scale(self, c):
        """
        Returns a new point that has been scaled according to the constant scaling factor c

        Args:
            c (float): The scaling factor

        Returns:
            Point: A point which coordinates have been scaled by the given factor
        """
        x1 = self.get_x() * c
        y1 = self.get_y() * c
        return Point(x1, y1)

    def dist(self, p):
        """
        Returns the distance between two points in the 2D-space. Calculated using the Euclidean norm.

        Args:
            p (Point): A Point object

        Returns:
            float: The distance between two points
        """
        d = math.pow((p.get_x() - self.get_x()), 2) + math.pow(
            (p.get_y() - self.get_y()), 2
        )
        return math.sqrt(d)



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


"""@file
FileHandler docstring

A collection of methods that can be used to write serialized protobuf packets to the binary version of the OBP-format.
"""

# Built-in

# PyPI

# Project


import json
from google.protobuf.internal.encoder import _VarintBytes
from google.protobuf.internal.decoder import _DecodeVarint32
import obplib.PacketHandler as PacketHandler
import obplib.OBP_pb2 as OBP
from obplib.Line import Line
from obplib.Curve import Curve
from obplib.TimedPoints import TimedPoints
from obplib.AcceleratingCurve import AcceleratingCurve
from obplib.AcceleratingLine import AcceleratingLine
from obplib.SyncPoint import SyncPoint


def write(obpdata, filename):
    """
    Writes the given data to a file.

    Args:
        obpdata (list): A list of OBP objects such as Line, TimedPoints and Curve
        filename (stream): The file to be written to
    """
    for i in range(0, len(obpdata)):
        data = obpdata[i].write_obp()
        filename.write(_VarintBytes(len(data)))
        filename.write(data)


def write_obp(obpdata, filename):
    """
    Writes obpData to a OBP-file with the given file name by calling Write(obpData, stream), if it exists. Creates a file and with the given name if it does not exist.

    Args:
        obpData (list): A list of OBP objects such as Line, TimedPoints and Curve
        filename (String): The name of the file to be written to
    """
    with open(filename, "wb+") as out:
        write(obpdata, out)
        out.flush()


def write_obpj(obpdata, filename):
    """
    Writes obpData to a OBPJ-file with the specified file name, if it exists. Creates a file with the given name if it does not exist.
    Args:
        obpData (list): A list OBP objects such as Line, TimedPoints and Curve
        filename (String): Name of the file to be written to
    """
    with open(filename, "w+") as out:
        out.write('{\n"OBP":[\n')
        lastind = len(obpdata) - 1

        for i in range(len(obpdata)):
            data = obpdata[i].get_obpj()
            check_keys(data)
            json.dump(data, out, indent=4)
            if i < lastind:
                out.write(",\n")
        out.write("\n]\n}\n")
        out.flush()


def read_obp(filename):
    """
    Reads a OBP-file with the specified file name and returns a list containing all OBP objects found wihin the file.

    Args:
        filename (String): Name of the file to be read from
    Returns:
            l (list): A list of OBP objects
    """
    l = []
    with open(filename, "rb") as f:
        data = f.read()
        next_pos, pos = 0, 0
        while pos < len(data):
            p = OBP.Packet()
            next_pos, pos = _DecodeVarint32(data, pos)
            p.ParseFromString(data[pos : pos + next_pos])
            pos = pos + next_pos
            temp = PacketHandler.handle_pkt(p)
            l.append(temp)
    f.close()
    return l


def read_obpj(filename):
    """
    Reads a OBPJ-file with the specified file name and returns a dictionary representation of the file in question

    Args:
        filename (String): obpj-file to be read
    """
    with open(filename, "r") as f:
        data = json.load(f)
        return data


def get_segment_from_dict(filename):
    """
    Method used by the compiler when reading obpj files. Reads a obpj file and returns a list of its constituent objects.

    Args:
        filename (String): obpj-file to be read
    Returns:
        l (list): A list of obp objects
    """
    l = []
    with open(filename, "r") as f:
        data = json.load(f)
        dicts = data["OBP"]
        for i in range(len(dicts)):
            k = dicts[i].keys()
            for key in k:
                if key == "line":
                    l.append(Line.from_dict(dicts[i][key]))
                if key == "curve":
                    l.append(Curve.from_dict(dicts[i][key]))
                if key == "timedPoints":
                    l.append(TimedPoints.from_dict(dicts[i][key]))
                if key == "acceleratingLine":
                    l.append(AcceleratingLine.from_dict(dicts[i][key]))
                if key == "acceleratingCurve":
                    l.append(AcceleratingCurve.from_dict(dicts[i][key]))
                if key == "syncPoint":
                    l.append(SyncPoint.from_dict(dicts[i][key]))
    f.close()
    return l


def check_keys(blob):
    """
    Checks that all keys are present before writing to obpj, used by both the compiler and the writer. Has to be done or else we won't be able
    to compile obpj files with default values. Checks each subdict in the file for missing keys and adds the key:value pair [key]:0 or [key]:False if it is not present.

    Needs refinement.

    Args:
        blob (dictionary): A dictionary
    """
    keys = ["x0", "y0", "x1", "y1"]
    for item in blob.keys():
        if item == "line" or item == "acceleratingLine":
            for key in keys:
                if key not in blob[item]:
                    blob[item][key] = 0

            if "beamPower" not in blob[item]["params"]:
                blob[item]["params"]["beamPower"] = 0
            if "spotSize" not in blob[item]["params"]:
                blob[item]["params"]["spotSize"] = 0

        if item == "timedPoints":
            if "beamPower" not in blob[item]["params"]:
                blob[item]["params"]["beamPower"] = 0
            if "spotSize" not in blob[item]["params"]:
                blob[item]["params"]["spotSize"] = 0

            current = blob[item]["points"][0]["t"]
            for i in range(len(blob[item]["points"])):
                if "t" in blob[item]["points"][i]:
                    current = blob[item]["points"][i]["t"]
                elif "t" not in blob[item]["points"][i]:
                    blob[item]["points"][i]["t"] = current

                if "x" not in blob[item]["points"][i]:
                    blob[item]["points"][i]["x"] = 0
                if "y" not in blob[item]["points"][i]:
                    blob[item]["points"][i]["y"] = 0

        if item == "syncPoint":
            if "duration" not in blob[item]:
                blob[item]["duration"] = 0
            if "value" not in blob[item]:
                blob[item]["value"] = False

        if item == "curve":
            points = ["p0", "p1", "p2", "p3"]
            for point in points:
                if "x" not in blob[item][point]:
                    blob[item][point]["x"] = 0
                if "y" not in blob[item][point]:
                    blob[item][point]["y"] = 0

            if "beamPower" not in blob[item]["params"]:
                blob[item]["params"]["beamPower"] = 0
            if "spotSize" not in blob[item]["params"]:
                blob[item]["params"]["spotSize"] = 0



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

import collections
import collections.abc

# This monkey-patch fixes this warning:
# /usr/local/lib/python3.9/dist-packages/svg/path/path.py:3:
# DeprecationWarning: Using or importing the ABCs from 'collections'
# instead of from 'collections.abc' is deprecated since Python 3.3,
# and in 3.10 it will stop working
collections.MutableSequence = collections.abc.MutableSequence



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# From PyPI
from google.protobuf.json_format import MessageToDict
import json

# Project
import obplib.OBP_pb2 as OBP
from obplib.Point import Point
from obplib.Beamparameters import Beamparameters


class AcceleratingCurve:
    """
    Constructor for the accelerating curve class

    Args:
        P1 (Point): The starting point
        P2 (Point): First control point
        P3 (Point): Second control point
        P4 (Point): End point
        speed (Int): Scan speed in µm/s
        bp (Beamparamters): Desired beam parameters
    """

    def __init__(self, P1, P2, P3, P4, si, sf, bp):
        self.P1 = P1
        self.P2 = P2
        self.P3 = P3
        self.P4 = P4
        self.si = si
        self.sf = sf
        self.bp = bp

    @classmethod
    def from_dict(cls, kw):
        """
        Recrates an instance of the accelerating curve class using a dictionary representation of said class. Assumes that that the given dictionary is in standard JSON structure.

        Args:
            kw: A dictionary representation of an accelerating curve
        Returns:
            cls: An instance of AcceleratingCurve recreated using kw
        """
        P1 = Point(kw["p0"]["x"], kw["p0"]["y"])
        P2 = Point(kw["p1"]["x"], kw["p1"]["y"])
        P3 = Point(kw["p2"]["x"], kw["p2"]["y"])
        P4 = Point(kw["p3"]["x"], kw["p3"]["y"])
        si = kw["si"]
        sf = kw["sf"]
        bp = Beamparameters(kw["params"]["spotSize"], kw["params"]["beamPower"])
        return cls(P1, P2, P3, P4, si, sf, bp)

    def get_pb(self):
        """
        Returns a protobuf packet containing curve information

        Returns:
            packet (binary): A protobuf packet containing curve information
        """

        pkt = OBP.Packet()
        accCurve = OBP.AcceleratingCurve()

        accCurve.params.CopyFrom(self.bp.get_pb())

        cp0 = OBP.Curve.Point()
        cp1 = OBP.Curve.Point()
        cp2 = OBP.Curve.Point()
        cp3 = OBP.Curve.Point()

        cp0.x = self.P1.get_x()
        cp0.y = self.P1.get_y()

        cp1.x = self.P2.get_x()
        cp1.y = self.P2.get_y()

        cp2.x = self.P3.get_x()
        cp2.y = self.P3.get_y()

        cp3.x = self.P4.get_x()
        cp3.y = self.P4.get_y()

        accCurve.p0.CopyFrom(cp0)
        accCurve.p1.CopyFrom(cp1)
        accCurve.p2.CopyFrom(cp2)
        accCurve.p3.CopyFrom(cp3)

        accCurve.si = int(self.si)
        accCurve.sf = int(self.sf)

        pkt.accelerating_curve.CopyFrom(accCurve)

        return pkt

    def get_packet(self, pkt):
        """
        Returns a serialize packet

        Args:
            packet (binary): A packet

        Returns:
            A serialized packet (String): A serialized packet
        """

        return pkt.SerializeToString()

    def write_obp(self):
        """
        Overloaded writer method - is called by the FileHandler when writing to obp

        Returns:
            serialized (binary): A serialized protobuf packet
        """
        return self.get_packet(self.get_pb())

    def get_obpj(self):
        """
        Overloaded writer method - is called by the FileHandler when writing to obpj

        Returns:
            d (dictionary): A dictionary representation of a protobuf packet
        """
        return MessageToDict(self.get_pb())

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def translate(self, v):
        """
        Returns a new AcceleratingCurve object which has been translated according to the given vector V

        Args:
            V (Vector): A translation vector

        Returns:
            AcceleratingCurve: An AcceleratingCurve which has been traslated according to the given vector
        """
        return AcceleratingCurve(
            self.P1.translate(v),
            self.P2.translate(v),
            self.P3.translate(v),
            self.P4.translate(v),
            self.si,
            self.sf,
            self.bp,
        )

    def rotate(self, theta):
        """
        Returns a new AcceleratingCurve which has been rotated theta radians around the origin

        Args:
            theta (radians): The amount of radians which the AcceleratingCurve is to be rotated

        Returns:
            AcceleratingCurve: An AcceleratingCurve rotated theta radians relative to the origin
        """
        return AcceleratingCurve(
            self.P1.rotate(theta),
            self.P2.rotate(theta),
            self.P3.rotate(theta),
            self.P4.rotate(theta),
            self.si,
            self.sf,
            self.bp,
        )

    def scale(self, factor):
        """
        Returns a new AcceleratingCurve scaled by the specified factor.

        Args:
            factor (float): The factor with which to scale the AcceleratingCurve

        Returns:
            AcceleratingCurve: An AcceleratingCurve scaled using the specified factor
        """
        return AcceleratingCurve(
            self.P1.scale(factor),
            self.P2.scale(factor),
            self.P3.scale(factor),
            self.P4.scale(factor),
            self.si,
            self.sf,
            self.bp,
        )



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


"""@SyncPoint docstring

    This class represents a digital output signal with a given name, a value(True/False), and a duration(ms).

"""

# PyPI
from google.protobuf.json_format import MessageToDict
import json

# Project
import obplib.OBP_pb2 as OBP


class SyncPoint:
    def __init__(self, endpoint, value, duration):
        """
        SyncPoint constructor

        Args:
            endpoint (String): The name to be used for the signal
            value (bool): The value of the signal
            duration (int): duration of the signal in milliseconds, this is the time during which the syncpoint will keep its given value. 0 = keep until reset.
        """
        self.endpoint = endpoint
        self.value = value
        self.duration = duration

    @classmethod
    def from_dict(cls, kw):
        """
        Recrates an instance of the SyncPoint class using a dictionary representation of said class. Assumes that that the given dictionary is in standard JSON structure.

        Args:
            kw: A dictionary representation of a SyncPoint
        Returns:
            cls: An instance of SyncPoint recreated using kw
        """
        endpoint = kw["endpoint"]
        value = kw["value"]
        duration = kw["duration"]

        return cls(endpoint, value, duration)

    def get_pb(self):
        """
        Returns a protobuf packet containing SyncPoint information

        Returns:
            Protobuf packet: A packet containing SyncPoint information
        """
        pkt = OBP.Packet()
        sp = OBP.SyncPoint()

        sp.endpoint = self.endpoint
        sp.value = self.value
        sp.duration = self.duration

        pkt.sync_point.CopyFrom(sp)
        return pkt

    def get_packet(self, pkt):
        """
        Returns a serialize packet

        Args:
            packet (binary): A packet

        Returns:
            A serialized packet (String): A serialized packet
        """
        return pkt.SerializeToString()

    def write_obp(self):
        """
        Overloaded writer method - is called by the FileHandler when writing to obp

        Returns:
            serialized (binary): A serialized protobuf packet
        """
        return self.get_packet(self.get_pb())

    def get_obpj(self):
        """
        Overloaded writer method - is called by the FileHandler when writing to obpj

        Returns:
            d (dictionary): A dictionary representation of a protobuf packet
        """
        d = MessageToDict(self.get_pb())
        return d

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def translate(self, V):
        pass

    def rotate(self, theta):
        pass

    def scale(self, factor):
        pass



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@TimedPoints docstring
A class containing a list of coordinate points, each with its own associated dwell time
"""

# PyPI
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message

import json

# Project
import obplib.OBP_pb2 as OBP
from obplib.Point import Point
from obplib.Beamparameters import Beamparameters
from obplib.PathSegment import PathSegment


class TimedPoints(PathSegment):
    def __init__(self, points, dwellTimes, bp):
        """
        Constructor for the TimedPoints class

        Args:
            points (list): A list of coordinate points
            dwellTimes (list): A list of dwell times
            bp (Beamparamters): Beam parameters

        Raises:
            Exception: Raised if points and dwellTimes are not of the same length
        """
        super().__init__()
        self.points = points
        self.dwellTimes = dwellTimes
        self.bp = bp
        if len(points) != len(dwellTimes):
            raise Exception("points and dwellTimes must be of the same length")

    @classmethod
    def from_dict(cls, kw):
        """
        Returns an instance of the TimedPoints class created using parameters specified in a dictionary

        Args:
            kw (dictionary): A dictionary representation of a TimedPoints object
        """
        points = kw["points"]
        p = []
        t = []
        for d in points:
            p.append(Point(d["x"], d["y"]))
            t.append(d["t"])
        bp = Beamparameters(kw["params"]["spotSize"], kw["params"]["beamPower"])
        return cls(p, t, bp)

    def get_packet(self, packet) -> bytes:
        """
        Returns a serialized string representation of a packet

        Args:
            packet (binary): A protobuf packet
        Returns:
            A serizalied packet (String): A serialized packet
        """
        pkt = packet.SerializeToString()
        return pkt

    def get_pb(self) -> Message:
        """
        Returns a protobuf packet containing TimedPoint information

        Returns:
           pkt (binary): A protobuf packet
        """
        pkt = OBP.Packet()
        timedPoints = OBP.TimedPoints()
        t = []
        timedPoints.params.CopyFrom(self.bp.get_pb())

        lastDwell = 0
        for i in range(len(self.points)):
            timedPoint = OBP.TimedPoints.TimedPoint()
            if self.dwellTimes[i] != lastDwell:

                lastDwell = self.dwellTimes[i]
                timedPoint.t = self.dwellTimes[i]
            p = Point(self.points[i].x, self.points[i].y)
            timedPoint.x = p.x
            timedPoint.y = p.y
            t.append(timedPoint)
            timedPoints.points.extend(t)
            t.clear()

        pkt.timed_points.CopyFrom(timedPoints)

        return pkt

    def write_obp(self) -> bytes:
        return self.get_packet(self.get_pb())

    def get_obpj(self) -> dict:
        """
        Overloaded writer method - is used by the FileHandler

        Returns:
            d (dictionary): A dictionary representation of a protobuf packet
        """
        return MessageToDict(self.get_pb())

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def translate(self, V):

        """
        Returns a TimedPoints object containing a list of points translated according to the given vector V

        Args:
            V (Vector): A translation vector

        Returns:
            TimedPoints: A TimedPoints object with a list of points which have been translated according to the given vector
        """
        pointlist = []
        dwell = []

        for i in range(0, len(self.points)):
            pointlist.append(self.points[i].translate(V))
            dwell.append(self.dwellTimes[i])

        return TimedPoints(pointlist, dwell, self.bp)

    def rotate(self, theta):
        """
        Returns a new TimedPoints object containing a list of points which have been rotated theta radians relative to the origin

        Args:
            theta (radians): The amount of radians which the points are to be rotated

        Returns:
            TimedPoints: A TimedPoints object containing a list of points that have been rotated theta radians relative to the origin

        """
        pointlist = []
        dwell = []

        for i in range(0, len(self.points)):
            pointlist.append(self.points[i].rotate(theta))
            dwell.append(self.dwellTimes[i])

        return TimedPoints(pointlist, dwell, self.bp)

    def scale(self, factor):
        """
        Returns a new TimedPoints object containing a list of points that have been scaled by the specified factor.

        Args:
            factor (float): The factor with which to scale the list of points

        Returns:
            Curve: A TimedPoints object containing a list of points scaled using the specified factor
        """

        pointlist = []
        dwell = []

        for i in range(0, len(self.points)):
            pointlist.append(self.points[i].scale(factor))
            dwell.append(self.dwellTimes[i])

        return TimedPoints(pointlist, dwell, self.bp)

    def get_segment_length(self) -> float:
        return 0

    def get_segment_duration(self) -> float:
        return sum(self.dwellTimes)



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

from .AcceleratingCurve import AcceleratingCurve
from .AcceleratingLine import AcceleratingLine
from .Beamparameters import Beamparameters
from .Curve import Curve
from .FileHandler import write_obp, write_obpj, read_obp, read_obpj
from .Line import Line
from .Point import Point
from .SyncPoint import SyncPoint
from .TimedPoints import TimedPoints
from .Transformations import translate, rotate, scale
from .Vector import Vector



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@Beamparameters docstring

This class represents parameters associated with a given beam segment.
"""

# Project
import obplib.OBP_pb2 as OBP


class Beamparameters:
    def __init__(self, spot_size, power):
        """
        Constructor for the Beamparameters class

        Args:
            spotSize (Int): Desired spot size in µm
            power (Int): Desired beam power in W
        """
        self.spot_size = spot_size
        self.power = power

    def get_pb(self):
        """
        Returns a protobuf packet containing the specified beam paramters

        Returns:
           Beamparameters packet (binary): A packet containing beam parameter information
        """
        pkt = OBP.BeamParameters()
        pkt.beam_power = self.power
        pkt.spot_size = self.spot_size
        return pkt



# SPDX-FileCopyrightText: 2022 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2023 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Built-in
import copy
import math

# Third-party
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import ipywidgets as widgets

# Project
import obplib as obp

mpl.style.use(["fast"])
plt.rcParams["figure.dpi"] = 400


class TimedPoint:
    """Helper class to store timed points."""

    def __init__(self, coords, dwell_time, bp):
        self.coords = coords
        self.dwell_time = dwell_time
        self.bp = bp


# Dictionary to define object types and their attributes
object_types = {
    obp.Line: {"speed_attr": "Speed", "point_attrs": ["P1", "P2"]},
    obp.Curve: {"speed_attr": "speed", "point_attrs": ["P1", "P2", "P3", "P4"]},
    obp.AcceleratingLine: {"speed_attr": ("si", "sf"), "point_attrs": ["p1", "p2"]},
    obp.AcceleratingCurve: {
        "speed_attr": ("si", "sf"),
        "point_attrs": ["P1", "P2", "P3", "P4"],
    },
    obp.TimedPoints: {},
    TimedPoint: {},
}


def unpack_timedpoints(timedpoints):
    """Unpack TimedPoints objects into a list of TimedPoint instances.

    Args:
        timedpoints (obp.TimedPoints): TimedPoints object.

    Returns:
        list: List of TimedPoint instances.
    """
    timedpoints_list = []
    dwell_time_const = False  # Flag to determine if dwell time is constant
    bp = timedpoints.bp
    if timedpoints.dwellTimes[1:] == [
        0 for _ in range(len(timedpoints.dwellTimes) - 1)
    ]:
        dwell_time_const = True
        dwell_time = timedpoints.dwellTimes[0]
    for i in range(len(timedpoints.points)):
        coords = [timedpoints.points[i].x, timedpoints.points[i].y]
        if not dwell_time_const:
            dwell_time = timedpoints.dwellTimes[i]
        timedpoints_list.append(TimedPoint(coords, dwell_time, bp))
    return timedpoints_list


def expand_objects(objects):
    """Expand objects into a flat list of objects by unpacking TimedPoints.

    Args:
        objects (list): List of objects.

    Returns:
        list: Expanded list of objects.
    """
    new_objects = []
    for obj in objects:
        obj_info = object_types.get(type(obj))
        if obj_info is None:
            continue  # Skip objects with unknown types
        if isinstance(obj, obp.TimedPoints):
            new_objects.extend(unpack_timedpoints(obj))
        else:
            new_objects.append(obj)
    return new_objects


def bezier_curve(control_points, num_points=100):
    """Generate a Bézier curve given control points.

    Args:
        control_points (list): List of control points.
        num_points (int, optional): Number of points on the curve.

    Returns:
        np.array: Array of points representing the Bézier curve.
    """
    t = np.linspace(0, 1, num_points)
    n = len(control_points) - 1
    curve = np.zeros((num_points, 2))
    for i in range(num_points):
        for j in range(n + 1):
            curve[i] += np.multiply(
                [control_points[j][0], control_points[j][1]],
                bernstein_polynomial(n, j, t[i]),
            )
    return curve


def binomial_coefficient(n, k):
    """Compute the binomial coefficient (n choose k).

    Args:
        n (int): Total number of items.
        k (int): Number of items to choose.

    Returns:
        float: Binomial coefficient value.
    """
    return math.factorial(n) / (math.factorial(k) * math.factorial(n - k))


def bernstein_polynomial(n, k, t):
    """Compute the Bernstein polynomial for a given n and k.

    Args:
        n (int): Degree of the polynomial.
        k (int): Index of the Bernstein polynomial.
        t (float): Parameter value.

    Returns:
        float: Value of the Bernstein polynomial.
    """
    return binomial_coefficient(n, k) * t**k * (1 - t) ** (n - k)


def mscatter(x, y, ax=None, m=None, **kw):
    """Create a scatter plot with custom markers. Taken from https://github.com/matplotlib/matplotlib/issues/11155#issuecomment-385939618

    Args:
        x (array-like): X-coordinates of the data points.
        y (array-like): Y-coordinates of the data points.
        ax (matplotlib.axes.Axes, optional): Matplotlib axis.
        m (list, optional): List of markers for data points.
        **kw: Additional keyword arguments for scatter.

    Returns:
        matplotlib.collections.PathCollection: Scatter plot object.
    """
    import matplotlib.markers as mmarkers

    if not ax:
        ax = plt.gca()
    sc = ax.scatter(x, y, **kw)
    if (m is not None) and (len(m) == len(x)):
        paths = []
        for marker in m:
            if isinstance(marker, mmarkers.MarkerStyle):
                marker_obj = marker
            else:
                marker_obj = mmarkers.MarkerStyle(marker)
            path = marker_obj.get_path().transformed(marker_obj.get_transform())
            paths.append(path)
        sc.set_paths(paths)
    return sc


def plot_lines_and_points(objects, upper_lim, object_number, show_control_points=False):
    """Plot lines and TimedPoints with color and size variation based on attributes.

    Args:
        objects (list): List of objects.
        upper_lim (int): Upper limit for the number of objects to plot.
        object_number (int): Number of objects to plot.
        show_control_points (bool, optional): Whether to display control points.

    Returns:
        None
    """
    fig, ax = plt.subplots()

    start_object_idx = max(upper_lim - object_number, 0)
    end_object_idx = min(upper_lim, len(objects))

    dwell_times_available = any(
        isinstance(obj, TimedPoint) for obj in objects[start_object_idx:end_object_idx]
    )
    line_speeds_available = any(
        isinstance(
            obj, (obp.Line, obp.AcceleratingLine, obp.Curve, obp.AcceleratingCurve)
        )
        for obj in objects[start_object_idx:end_object_idx]
    )

    if dwell_times_available:
        cmap_dwell_times = plt.get_cmap(
            "viridis"
        )  # Colormap for TimedPoints' dwellTimes
        dwell_times_norm = None
        dwell_times_values = [
            obj.dwell_time for obj in objects
        ]  # Collect all dwell time values

        # Check if there is only one unique value for dwell times
        unique_dwell_times = np.unique(dwell_times_values)
        if len(unique_dwell_times) == 1:
            # If there is only one unique value, manually adjust the normalization
            dwell_times_norm = mcolors.Normalize(
                unique_dwell_times[0] * 0.9, unique_dwell_times[0] * 1.1
            )
        else:
            dwell_times_norm = mcolors.Normalize(
                np.min(dwell_times_values), np.max(dwell_times_values)
            )
    if line_speeds_available:
        cmap_line_speeds = plt.get_cmap(
            "Wistia"
        )  # Colormap for line speed/acceleration
        line_speeds_norm = None

    scatter_points_timed = []  # Separate scatter points list for TimedPoints
    scatter_points_line = []  # Separate scatter points list for line objects
    scatter_sizes_timed = []
    scatter_sizes_line = []
    markers_timed = []
    markers_line = []

    scatter_colors_dwell_times = []  # List for TimedPoints' dwellTimes
    scatter_colors_line_speeds = []  # List for line speed/acceleration

    for i in range(start_object_idx, end_object_idx):
        last_object = i == end_object_idx - 1
        obj = objects[i]

        obj_info = object_types.get(type(obj))
        if obj_info is None:
            continue  # Skip objects with unknown types

        if isinstance(obj, TimedPoint):
            scatter_points_timed.append([obj.coords[0], obj.coords[1]])
            if dwell_times_available:
                scatter_colors_dwell_times.append(obj.dwell_time)
            if last_object:
                markers_timed.append("*")
                scatter_sizes_timed.append(obj.bp.spot_size * 0.3)
            else:
                markers_timed.append("o")
                scatter_sizes_timed.append(obj.bp.spot_size * 0.01)

        elif isinstance(
            obj, (obp.Line, obp.AcceleratingLine, obp.Curve, obp.AcceleratingCurve)
        ):
            scatter_params = plot_line(
                ax, obj, cmap_line_speeds, show_control_points=show_control_points
            )
            if last_object:
                last_point = scatter_params[0]
                scatter_points_line.append([last_point[1][0], last_point[1][1]])
                scatter_sizes_line.append(scatter_params[1])
                markers_line.append("*")
            if len(scatter_params) > 3:
                for i in range(len(scatter_params[2])):
                    scatter_points_line.append(
                        [scatter_params[2][0][i][0], scatter_params[2][0][i][1]]
                    )
                    scatter_sizes_line.append(scatter_params[1] * 0.3)
                    markers_line.append(scatter_params[2][1])
            scatter_colors_line_speeds.extend(scatter_params[-1])

    scatter_points_timed = np.array(scatter_points_timed)
    scatter_points_line = np.array(scatter_points_line)

    if dwell_times_available:
        # Check if there is only one unique value for dwell times
        unique_dwell_times = np.unique(dwell_times_values)
        if len(unique_dwell_times) == 1:
            # If there is only one unique value, manually adjust the normalization
            dwell_times_norm = mcolors.Normalize(
                unique_dwell_times[0] * 0.9, unique_dwell_times[0] * 1.1
            )
        else:
            dwell_times_norm = mcolors.Normalize(
                np.min(dwell_times_values), np.max(dwell_times_values)
            )

        last_point_timed = scatter_points_timed[-1]
        # Plot all TimedPoints except the last one with the specified parameters

        dwell_times_norm = mcolors.Normalize(
            np.min(scatter_colors_dwell_times), np.max(scatter_colors_dwell_times)
        )
        dwell_times_cbar = plt.colorbar(
            plt.cm.ScalarMappable(norm=dwell_times_norm, cmap=cmap_dwell_times), ax=ax
        )
        dwell_times_cbar.set_label("Dwell Time", rotation=270, labelpad=20)
        plt.scatter(
            scatter_points_timed[:-1, 0],
            scatter_points_timed[:-1, 1],
            c=scatter_colors_dwell_times[:-1],
            s=scatter_sizes_timed[:-1],
            marker="o",
            zorder=100,
        )
        plt.scatter(
            last_point_timed[0],
            last_point_timed[1],
            c="white",
            marker="*",
            s=scatter_sizes_timed[-1],
            zorder=101,
        )
    if line_speeds_available:
        last_point_line = scatter_points_line[-1]
        line_speeds_norm = mcolors.Normalize(
            np.min(scatter_colors_line_speeds), np.max(scatter_colors_line_speeds)
        )
        line_speeds_cbar = plt.colorbar(
            plt.cm.ScalarMappable(norm=line_speeds_norm, cmap=cmap_line_speeds),
            ax=ax,
            label="Speed",
            location="left",
            ticklocation="left",
        )
        # Plot all line objects except the last one with the specified parameters
        if len(scatter_points_line) > 1:
            plt.scatter(
                scatter_points_line[:-1, 0],
                scatter_points_line[:-1, 1],
                c="white",
                s=scatter_sizes_line[:-1],
                marker="*",
                zorder=100,
            )
        # Plot the last line object with different parameters (white star)
        mscatter(
            last_point_line[0],
            last_point_line[1],
            c="white",
            marker=markers_line[-1],
            s=scatter_sizes_line[-1],
            zorder=101,
        )

    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=45, ha="right")
    ax.set_facecolor("xkcd:black")
    ax.set_aspect("equal")
    ax.autoscale(enable=False)
    # Add vertical and horizontal lines for the axes
    ax.axvline(
        x=0, color="white", linewidth=0.5, linestyle="-", clip_on=False, zorder=0
    )
    ax.axhline(
        y=0, color="white", linewidth=0.5, linestyle="-", clip_on=False, zorder=0
    )
    plt.grid(False)
    plt.show()


def plot_line(ax, line, cmap, show_control_points=False):
    """Plot a line with color and size variation based on speed and spot size.

    Args:
        ax (matplotlib.axes.Axes): Matplotlib axis to plot on.
        line: Line object to plot.
        cmap (matplotlib.colors.Colormap): Colormap for color variation.
        show_control_points (bool, optional): Whether to display control points.

    Returns:
        list: List containing plot parameters.
    """
    num_segments = 100

    obj_info = object_types.get(type(line))
    if obj_info is None:
        return

    speed_attr = obj_info["speed_attr"]
    point_attrs = obj_info["point_attrs"]

    control_points = np.array(
        [
            [
                copy.copy(getattr(line, point_attr).x),
                copy.copy(getattr(line, point_attr).y),
            ]
            for point_attr in point_attrs
        ]
    )

    curve = bezier_curve(control_points, num_segments)

    if isinstance(speed_attr, tuple):
        speed_attr_value = (getattr(line, speed_attr[0]), getattr(line, speed_attr[1]))
    else:
        speed_attr_value = getattr(line, speed_attr, None)
    if isinstance(speed_attr_value, tuple):
        speeds = np.linspace(speed_attr_value[0], speed_attr_value[1], num_segments - 1)
    else:
        speeds = np.linspace(speed_attr_value, speed_attr_value, num_segments - 1)

    spot_sizes = np.linspace(line.bp.spot_size, line.bp.spot_size, num_segments - 1)

    curve_array = np.stack([curve[:-1], curve[1:]], axis=1).reshape(-1, 2, 2)
    lc = mpl.collections.LineCollection(curve_array, cmap=cmap, zorder=80)
    lc.set_array(np.array(speeds))
    lc.set_linewidth(spot_sizes * 0.01)
    ax.add_collection(lc)
    if show_control_points:
        t = mpl.markers.MarkerStyle(marker="*")
        t._transform = t.get_transform().rotate_deg(180)
        control_points = np.array(
            [
                [
                    copy.copy(getattr(line, point_attr).x),
                    copy.copy(getattr(line, point_attr).y),
                ]
                for point_attr in point_attrs
            ]
        )
        control_params = [control_points, t]
    if show_control_points:
        return [curve_array[-1], spot_sizes[-1], control_params, speeds]
    else:
        return [curve_array[-1], spot_sizes[-1], speeds]


def notebook_viewer(objects, show_control_points=False):
    """Create an interactive viewer for lines and curves.

    Args:
        objects (list): List of objects to display.
        show_control_points (bool, optional): Whether to show control points.

    Returns:
        None
    """
    objects = expand_objects(objects)
    object_number = widgets.IntText(
        value=len(objects), description="Max Objects:", min=1, max=len(objects)
    )
    widgets.interact(
        plot_lines_and_points,
        objects=widgets.fixed(objects),
        upper_lim=(1, len(objects), 1),
        description="Index of objects",
        object_number=object_number,
        show_control_points=widgets.fixed(show_control_points),
        layout=widgets.Layout(width="50%"),
        continuous_update=False,
    )


# TODO fix the colormap showing the lowermost value of the colorbar when there is only one value in the colormap
# TODO fix the control points so they can be shown
# TODO tentative: interpolate the segments of the curve for increased smoothness?



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

# Built-in
from xml.dom import minidom

# PyPI
from svg.path import Line, CubicBezier
from svg.path import parse_path

# Project
from . import segment_handler


def read_svg(f):
    doc = minidom.parse(f)
    width, height = int(doc.getElementsByTagName("svg")[0].getAttribute("width")), int(
        doc.getElementsByTagName("svg")[0].getAttribute("height")
    )
    paths = get_paths(doc)
    doc.unlink()
    return paths, width, height


def get_rects(doc):
    rects = []
    for path in doc.getElementsByTagName("rect"):
        recdict = {}
        recdict["x"] = float(path.getAttribute("x"))
        recdict["y"] = float(path.getAttribute("y"))
        recdict["width"] = float(path.getAttribute("width"))
        recdict["height"] = float(path.getAttribute("height"))
        rects.append(recdict)
    return rects


def get_paths(doc):
    path_strings = [path.getAttribute("d") for path in doc.getElementsByTagName("path")]
    return path_strings


def handle_elements(element, width, height):
    l = []
    for el in element:
        for path_element in parse_path(el):
            if isinstance(path_element, Line):
                l.append(
                    segment_handler.create_segment("Line", path_element, width, height)
                )
            elif isinstance(path_element, CubicBezier):
                l.append(
                    (
                        segment_handler.create_segment(
                            "Curve", path_element, width, height
                        )
                    )
                )

    return l


def convert_svg(element, width, height):
    l = handle_elements(element, width, height)
    l1 = [x for x in l if x != []]
    return l1



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@Transformations docstring
A collection of methods to be used when dealing with transformations of lists of beam segments.

"""


def translate(obpdata, V):
    translated = []
    for segment in obpdata:
        translated.append(segment.translate(V))
    return translated


def rotate(obpdata, theta):
    rotated = []
    for segment in obpdata:
        rotated.append(segment.rotate(theta))
    return rotated


def scale(obpdata, factor):
    scaled = []
    for segment in obpdata:
        scaled.append(segment.scale(factor))
    return scaled



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@file
pktHandler docstring

A collection of methods used when handling protocol buffer pkts
"""

# Project




from obplib.Line import Line
from obplib.Point import Point
from obplib.TimedPoints import TimedPoints
from obplib.Beamparameters import Beamparameters
from obplib.Curve import Curve
from obplib.AcceleratingLine import AcceleratingLine
from obplib.AcceleratingCurve import AcceleratingCurve
from obplib.SyncPoint import SyncPoint
def handle_line(pkt):
    """
    Handles pkts containing line objects

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        l (Line): A OBP line object
    """
    p1 = Point(pkt.line.x0, pkt.line.y0)
    p2 = Point(pkt.line.x1, pkt.line.y1)
    bp = pkt.line.params
    bp1 = Beamparameters(bp.spot_size, bp.beam_power)
    speed = pkt.line.speed
    l = Line(p1, p2, speed, bp1)
    return l


def handle_timed_points(pkt):
    """
    Handles pkts containing Timed Points objects

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        tp (TimedPoints): A OBP TimedPoints object
    """
    points = pkt.timed_points.points
    dwellTimes = []
    obp_points = []

    for i in range(len(points)):
        dwellTimes.append(points[i].t)
        obp_points.append(Point(points[i].x, points[i].y))

    bp = pkt.timed_points.params
    bp1 = Beamparameters(bp.spot_size, bp.beam_power)
    tp = TimedPoints(obp_points, dwellTimes, bp1)
    return tp


def handle_curve(pkt):
    """
    Handles pkts containing Curve objects

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        c (Curve): A OBP Curve object
    """
    p0 = Point(pkt.curve.p0.x, pkt.curve.p0.y)
    p1 = Point(pkt.curve.p1.x, pkt.curve.p1.y)
    p2 = Point(pkt.curve.p2.x, pkt.curve.p2.y)
    p3 = Point(pkt.curve.p3.x, pkt.curve.p3.y)
    speed = pkt.curve.speed
    bp = pkt.curve.params
    bp1 = Beamparameters(bp.spot_size, bp.beam_power)

    return Curve(p0, p1, p2, p3, speed, bp1)


def handle_accelerating_line(pkt):
    """
    Handles pkts containing Accelerating Line objects

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        AcceleratingLine (Accelerating line): An accelerating line
    """
    p1 = Point(pkt.accelerating_line.x0, pkt.accelerating_line.y0)
    p2 = Point(pkt.accelerating_line.x1, pkt.accelerating_line.y1)
    v0 = pkt.accelerating_line.si
    v1 = pkt.accelerating_line.sf
    bp = Beamparameters(
        pkt.accelerating_line.params.spot_size,
        pkt.accelerating_line.params.beam_power,
    )
    return AcceleratingLine(p1, p2, v0, v1, bp)


def handle_sync_point(pkt):
    """
    Handles pkts containing sync point information

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        SyncPoint (Sync point): A SyncPoint object
    """
    endpoint = pkt.sync_point.endpoint
    value = pkt.sync_point.value
    duration = pkt.sync_point.duration

    return SyncPoint(endpoint, value, duration)


def handle_accelerating_curve(pkt):
    """
    Handles pkts containing accelerating curves

    Args:
        pkt (protobuf): A protobuf pkt
    Returns:
        AcceleratingCurve (accelerating curve): A AcceleratingCurve object
    """
    p1 = Point(pkt.accelerating_curve.p0.x, pkt.accelerating_curve.p0.y)
    p2 = Point(pkt.accelerating_curve.p1.x, pkt.accelerating_curve.p1.y)
    p3 = Point(pkt.accelerating_curve.p2.x, pkt.accelerating_curve.p2.y)
    p4 = Point(pkt.accelerating_curve.p3.x, pkt.accelerating_curve.p3.y)

    si = pkt.accelerating_curve.si
    sf = pkt.accelerating_curve.sf

    bp = Beamparameters(
        pkt.accelerating_curve.params.spotSize, pkt.accelerating_curve.params.power
    )

    return AcceleratingCurve(p1, p2, p3, p4, si, sf, bp)


def handle_pkt(pkt):
    """
    This methods is responsible for the delegation of pkt handling
    dependent upon the contents of said pkt

    Args:
        pkt (protobuf): A protobuf pkt

    Returns:
        A pkt handler: Delegated pkt handler
    """
    if pkt.HasField("line"):
        return handle_line(pkt)
    elif pkt.HasField("timed_points"):
        return handle_timed_points(pkt)
    elif pkt.HasField("curve"):
        return handle_curve(pkt)
    elif pkt.HasField("accelerating_line"):
        return handle_accelerating_line(pkt)
    elif pkt.HasField("accelerating_curve"):
        return handle_accelerating_curve(pkt)
    elif pkt.HasField("sync_point"):
        return handle_sync_point(pkt)
    else:
        print(pkt)
        raise Exception("Unknown pkt recieved")



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

# Built-in
import math
import logging

# Freemelt
from ..Line import Line
from ..Curve import Curve
from ..TimedPoints import TimedPoints
from ..Point import Point
from ..Beamparameters import Beamparameters


log = logging.getLogger(__name__)


def create_segment(t, element, width, height):

    xscale = 100000 / width
    yscale = 100000 / height
    xoffset = -width / 2
    yoffset = -height / 2

    def conv_x(x):
        return float((x + xoffset) * xscale)

    def conv_y(y):
        return float((y + yoffset) * yscale)

    if t == "Line":
        p1 = Point(conv_x(element.start.real), conv_y(element.start.imag))
        p2 = Point(conv_x(element.end.real), conv_y(element.end.imag))
        l = Line(p1, p2, 100000, Beamparameters(100, 100))
    elif t == "Curve":
        p1 = Point(conv_x(element.start.real), conv_y(element.start.imag))
        p2 = Point(conv_x(element.control1.real), conv_y(element.control1.imag))
        p3 = Point(conv_x(element.control2.real), conv_y(element.control2.imag))
        p4 = Point(conv_x(element.end.real), conv_y(element.end.imag))
        l = Curve(p1, p2, p3, p4, 100000, Beamparameters(100, 100))
    return l


# Work in progress
def validate_segment(seg):
    if isinstance(seg, Line):
        length = math.sqrt(
            math.pow(seg.P2.X - seg.P1.X, 2) + math.pow(seg.P2.Y - seg.P1.Y, 2)
        )
        if length > 100000:
            log.warning("Max length of line exceeded")
        elif seg.Speed == 0:
            log.warning("One or more lines has a speed of 0")
    if isinstance(seg, Curve):
        length = math.sqrt(
            math.pow(seg.P4.X - seg.P1.X, 2) + math.pow(seg.P4.Y - seg.P1.Y, 2)
        )
        if length > 100000:
            log.warning("Curve segment longer than 10cm")
        elif seg.speed == 0:
            log.warning("One or more curves has a speed of 0")
    if isinstance(seg, TimedPoints):
        for point in seg.points:
            if point.X or point.Y > 100000:
                log.warning("A point is located outside of the build area")



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


"""@Vector docstring

A class representing a 2 or 3-dimensional vector. The constructor default to 2D if not z-coordinate is given.
"""


class Vector:
    def __init__(self, x, y, z=None):
        """
        Constructor for the Vector class

        Args:
            x (Int): x-coordinate
            y (Int): y-coordinate
            z (Int, optional): z-coordinate. Defaults to None.
        """
        self.X = x
        self.Y = y
        self.Z = z

    def get_x(self):
        return self.X

    def get_y(self):
        return self.Y

    def get_z(self):
        return self.Z

    def add(self, V):
        """
        Adds two vectors and returns as new vector which elements are the element-wise sum of the two others

        Args:
            V (Vector): A 2 or 3-dimensional vector

        Returns:
            Vector: A vector which elements are the element-wise sum of the two others
        """
        return Vector(
            self.get_x() + V.get_x(), self.get_y() + V.get_y(), self.get_z() + V.get_Z()
        )

    def subtract(self, V):
        """
        Subtracts two vectors and returns as new vector which elements are the element-wise difference of the two others

        Args:
            V (Vector): A 2 or 3-dimensional vector

        Returns:
            Vector: A vector which elements are the element-wise difference of the two others
        """
        return Vector(
            self.get_x() - V.get_x(), self.get_y() - V.get_y(), self.get_z() - V.get_z()
        )

    def dot_prod(self, V):
        """
        Returns the dot product of two vectors

        Args:
            V (Vector): A 2 or 3-dimensional vector

        Returns:
            int: The dot product of two vectors
        """
        return (
            self.get_x() * V.get_x()
            + self.get_y() * V.get_y()
            + self.get_z() * V.get_z()
        )



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


"""@Line docstring
A class which describes a line drawn between point P1 and P2 using a specified scan speed and specified beam paramters
"""

# PyPI
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
import math

import json


# Project
import obplib.OBP_pb2 as OBP
from obplib.Point import Point
from obplib.Beamparameters import Beamparameters
from obplib.PathSegment import PathSegment


class Line(PathSegment):
    def __init__(self, P1, P2, Speed, bp):
        """
        Constructor for the Line class

        Args:
            P1 (Point): The start point
            P2 (Point): The end point
            Speed (Int): Speed in µm/s
            bp (Beamparameters): Beam parameters
        """
        super().__init__()
        self.P1 = P1
        self.P2 = P2
        self.Speed = Speed
        self.bp = bp

    @classmethod
    def from_dict(cls, kw):
        """
        Returns an instance of the Line class created using parameters specified in a dictionary

        Args:
            kw (dictionary): A dictionary representation of a Line object
        """
        P1 = Point(kw["x0"], kw["y0"])
        P2 = Point(kw["x1"], kw["y1"])
        speed = int(kw["speed"])
        bp = Beamparameters(kw["params"]["spotSize"], kw["params"]["beamPower"])
        return cls(P1, P2, speed, bp)

    def get_pb(self) -> Message:
        """
        Returns a protobuf packet containing Line information

        Returns:
            pkt (binary): A packet protobuf containing line info
        """
        pkt = OBP.Packet()
        line1 = OBP.Line()
        line1.params.CopyFrom(self.bp.get_pb())
        line1.x0 = self.P1.x
        line1.x1 = self.P2.x
        line1.y0 = self.P1.y
        line1.y1 = self.P2.y
        line1.speed = self.Speed
        pkt.line.CopyFrom(line1)

        return pkt

    def get_packet(self, packet) -> bytes:
        """
        Returns a serialized packet

        Args:
            packet (binary): A packet

        Returns:
            A serialized packet (String): A serialized packet
        """
        pkt = packet.SerializeToString()
        return pkt

    def write_obp(self) -> bytes:
        """
        Overloaded writer method - is called by the FileHandler when writing to obp

        Returns:
            serialized (binary): A serialized protobuf packet
        """
        return self.get_packet(self.get_pb())

    def get_obpj(self) -> dict:
        """
        Overloaded writer method - is called by the FileHandler when writing to obpj

        Returns:
            d (dictionary): A dictionary representation of a protobuf packet
        """
        d = MessageToDict(self.get_pb())
        return d

    def translate(self, V):
        """
        Returns a new Line object which has been translated according to the given vector V

        Args:
            V (Vector): A translation vector

        Returns:
            Line: A Line which has been traslated according to the given vector
        """
        return Line(self.P1.translate(V), self.P2.translate(V), self.Speed, self.bp)

    def rotate(self, theta):
        """
        Returns a new Line which has been rotated theta radians around the origin

        Args:
            theta (radians): The amount of radians which the Line is to be rotated

        Returns:
            Line: A Line rotated theta radians relative to the origin
        """
        return Line(self.P1.rotate(theta), self.P2.rotate(theta), self.Speed, self.bp)

    def scale(self, factor):
        """
        Returns a new Line scaled by the specified factor.

        Args:
            factor (float): The factor with which to scale the Line

        Returns:
            Line: A Line scaled using the specified factor
        """
        return Line(self.P1.scale(factor), self.P2.scale(factor), self.Speed, self.bp)

    def get_segment_length(self) -> float:
        return math.sqrt(
            math.pow(self.P2.get_x() - self.P1.get_x(), 2)
            + math.pow(self.P2.get_y() - self.P1.get_y(), 2)
        )

    def get_segment_duration(self) -> float:
        return self.get_segment_length() / self.Speed

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def null_attr(self, attr):
        if (attr) == None:
            return 0



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@AcceleratingLine docstring
A class describing a straigh line drawn between points p1 and p2, with initial speed v0 and final speed v1.
"""

# PyPI
from google.protobuf.json_format import MessageToDict
import json

# Project
import obplib.OBP_pb2 as OBP
from obplib.Point import Point
from obplib.Beamparameters import Beamparameters


class AcceleratingLine:
    def __init__(self, p1, p2, si, sf, bp):
        """
        Constructor for the AcceleratingLine class

        Args:
            p1 (Point): The start point
            p2 (Point): The end point
            v0 (Int): Initial speed in µm/s
            v1 (Int): Final speed in µm/s
            bp (Beamparameters): Beam parameters
        """
        self.si = si
        self.sf = sf
        self.p1 = p1
        self.p2 = p2
        self.bp = bp

    @classmethod
    def from_dict(cls, kw):
        """
        Returns an instance of the AcceleratingLine class created using parameters specified in a dictionary

        Args:
            kw (dictionary): A dictionary representation of an AcceleratingLine object
        """
        v0 = kw["si"]
        v1 = kw["sf"]
        p0 = Point(kw["x0"], kw["y0"])
        p1 = Point(kw["x1"], kw["y1"])
        bp = Beamparameters(kw["params"]["spotSize"], kw["params"]["beamPower"])
        return cls(p0, p1, v0, v1, bp)

    def get_pb(self):
        """
        Returns a protobuf packet containing AcceleratingLine information

        Returns:
            Protobuf packet: A packet containing AccelerationLine info
        """
        pkt = OBP.Packet()
        acc_line = OBP.AcceleratingLine()

        acc_line.params.CopyFrom(self.bp.get_pb())

        acc_line.x0 = self.p1.get_x()
        acc_line.y0 = self.p1.get_y()
        acc_line.x1 = self.p2.get_x()
        acc_line.y1 = self.p2.get_y()

        acc_line.si = int(self.si)
        acc_line.sf = int(self.sf)

        pkt.accelerating_line.CopyFrom(acc_line)

        return pkt

    def get_packet(self, pkt):
        """
        Returns a serialized packet
        Args:
            pkt (binary): A packet

        Returns:
            A serialized packet (String): A serialized packet
        """
        return pkt.SerializeToString()

    def write_obp(self):
        return self.get_packet(self.get_pb())

    def get_obpj(self):
        """
        Overloaded writer method - is used by the FileHandler

        Returns:
            d (dictionary): A dictionary representation of a protobuf packet
        """
        return MessageToDict(self.get_pb())

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def translate(self, V):
        """
        Returns a new AcceleratomgLine object which has been translated according to the given vector V

        Args:
            V (Vector): A translation vector

        Returns:
            AcceleratingLine: An AcceleratingLine which has been traslated according to the given vector
        """
        return AcceleratingLine(
            self.p1.translate(V), self.p2.translate(V), self.si, self.sf, self.bp
        )

    def rotate(self, theta):
        """
        Returns a new AcceleratingLine which has been rotated theta radians around the origin

        Args:
            theta (radians): The amount of radians which the AccelerartingLine is to be rotated

        Returns:
            AcceleratingLine: An AcceleratingLine rotated theta radians relative to the origin
        """
        return AcceleratingLine(
            self.p1.rotate(theta), self.p2.rotate(theta), self.si, self.sf, self.bp
        )

    def scale(self, factor):
        """
        Returns a new AcceleratingLine scaled by the specified factor.

        Args:
            factor (float): The factor with which to scale the AcceleratingLine

        Returns:
            AcceleratingLine: An AcceleratingLine scaled using the specified factor
        """
        return AcceleratingLine(
            self.p1.scale(factor), self.p2.scale(factor), self.si, self.sf, self.bp
        )



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

# Built-in
import sys

# PyPI
import setuptools

from setuptools.command.build_py import build_py
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info

import pkg_resources

with open("README.md", "r") as fh:
    description = fh.read()

sys.path.append("obplib")
from _version import __version__

sys.path.remove("obplib")


class Protoc(setuptools.Command):
    """Use protoc to generate OBP_pb2.py and OBP_pb2_grpc.py"""

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import grpc_tools.protoc

        # Explicitly include the built-in proto files from grpc_tools
        # because OBP.proto uses google.protobuf.Any. Seems to be
        # necessary on Windows only.
        proto_include = pkg_resources.resource_filename("grpc_tools", "_proto")
        command = [
            "grpc_tools.protoc",
            "-I=.",
            f"-I={proto_include}",
            "--python_out=.",
            "--grpc_python_out=.",
            "obplib/OBP.proto",
        ]
        code = grpc_tools.protoc.main(command)
        command_str = " ".join(command)
        assert code == 0, f"Command {command_str!r} failed with exit code {code}"


# python3 setup.py build
class BuildCommand(build_py):
    def run(self):
        self.run_command("protoc")
        super().run()


# python3 setup.py bdist_wheel
# pip will call "python3 setup.py bdist_wheel" if 'wheel' is installed,
# otherwise it will fallback to "python setup.py install".
try:
    from wheel.bdist_wheel import bdist_wheel

    class WheelCommand(bdist_wheel):
        def run(self):
            self.run_command("protoc")
            super().run()


except ModuleNotFoundError:
    WheelCommand = None


# python3 setup.py install
class InstallCommand(install):
    def run(self):
        self.run_command("protoc")
        super().run()


# python3 setup.py develop
class DevelopCommand(develop):
    def run(self):
        self.run_command("protoc")
        super().run()


setuptools.setup(
    name="obplib",
    version=__version__,
    license="apache-2.0",
    author="Freemelt AB",
    author_email="opensource@freemelt.com",
    description="A library for the creation of beam paths",
    long_description=description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/freemelt/openmelt/obplib-python",
    download_url="https://gitlab.com/freemelt/openmelt/obplib-python/-/archive/0.3.0/obplib-python-0.3.0.tar.gz",
    keywords="obp openbeampath freemelt",
    packages=setuptools.find_packages(exclude=["tests"]),
    entry_points={
        "console_scripts": [
            "obpc=obplib.compiler.__main__:OBPC",
            "obpviewer=obplib.viewer.obpviewer:main",
        ]
    },
    setup_requires=[
        "grpcio-tools",
    ],
    install_requires=[
        "protobuf",
        "grpcio",
        "svg.path==3.0",
        "click",
    ],
    cmdclass={
        "build_py": BuildCommand,
        "install": InstallCommand,
        "develop": DevelopCommand,
        # "egg_info": EggInfoCommand,
        "bdist_wheel": WheelCommand,
        "protoc": Protoc,
    },
    package_data={"obplib": ["OBP.proto"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



# SPDX-FileCopyrightText: 2019,2020 Freemelt AB
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright 2019,2020 Freemelt AB

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""@Curve docstring
A class which describes a cubic bezier curve corresponding to a given speed and given parameters.
The curve starts at P1, aims at P2 and ends at P4, via P3.
"""

# PyPI
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
import math


import json


# Project
import obplib.OBP_pb2 as OBP
from obplib.Beamparameters import Beamparameters
from obplib.Point import Point
from obplib.PathSegment import PathSegment


class Curve(PathSegment):
    def __init__(self, P1, P2, P3, P4, speed, bp):
        """

        Constructor for the Curve class

        Args:
           P1 (Point): The starting point
           P2 (Point): First control point
           P3 (Point): Second control point
           P4 (Point): End point
           speed (Int): Scan speed in µm/s
           bp (Beamparamters): Desired beam parameters
        """
        super().__init__()
        self.P1 = P1
        self.P2 = P2
        self.P3 = P3
        self.P4 = P4
        self.speed = speed
        self.bp = bp

    @classmethod
    def from_dict(cls, kw):
        """
        Returns an instance of the Curve class created using parameters specified in a dictionary

        Args:
            kw (dictionary): A dictionary representation of a Curve object
        """
        P1 = Point(kw["p0"]["x"], kw["p0"]["y"])
        P2 = Point(kw["p1"]["x"], kw["p1"]["y"])
        P3 = Point(kw["p2"]["x"], kw["p2"]["y"])
        P4 = Point(kw["p3"]["x"], kw["p3"]["y"])
        bp = Beamparameters(kw["params"]["spotSize"], kw["params"]["beamPower"])
        speed = int(kw["speed"])
        return cls(P1, P2, P3, P4, speed, bp)

    def get_pb(self) -> Message:
        """
        Returns a protobuf packet containing curve information

        Returns:
            pkt (binary): A protobuf packet containing curve information
        """
        pkt = OBP.Packet()
        curve = OBP.Curve()

        curve.speed = self.speed
        curve.params.CopyFrom(self.bp.get_pb())

        cp0 = OBP.Curve.Point()
        cp1 = OBP.Curve.Point()
        cp2 = OBP.Curve.Point()
        cp3 = OBP.Curve.Point()

        cp0.x = self.P1.get_x()
        cp0.y = self.P1.get_y()

        cp1.x = self.P2.get_x()
        cp1.y = self.P2.get_y()

        cp2.x = self.P3.get_x()
        cp2.y = self.P3.get_y()

        cp3.x = self.P4.get_x()
        cp3.y = self.P4.get_y()

        curve.p0.CopyFrom(cp0)
        curve.p1.CopyFrom(cp1)
        curve.p2.CopyFrom(cp2)
        curve.p3.CopyFrom(cp3)

        pkt.curve.CopyFrom(curve)

        return pkt

    def get_packet(self, packet) -> bytes:
        """
        Returns a serialized packet
        Args:
            packet (binary): A packet

        Returns:
            A serialized packet (String): A serialized packet
        """
        pkt = packet.SerializeToString()
        return pkt

    def write_obp(self) -> bytes:
        return self.get_packet(self.get_pb())

    def get_obpj(self) -> dict:
        """
        Overloaded writer method - is used by the FileHandler

        Returns:
             (dictionary): A dictionary representation of a protobuf packet
        """
        return MessageToDict(self.get_pb())

    def __repr__(self):
        return json.dumps(self.get_obpj(), indent=2)

    def translate(self, V):
        """
        Returns a new curve that has been translated using the translation vector V

        Args:
            V (Vector): A translation vector

        Returns:
            Curve: A curve that has been translate according to the given vector
        """
        return Curve(
            self.P1.translate(V),
            self.P2.translate(V),
            self.P3.translate(V),
            self.P4.translate(V),
            self.speed,
            self.bp,
        )

    def rotate(self, theta):
        """
        Returns a new Curve which has been rotated theta radians around the origin

        Args:
            theta (radians): The amount of radians which the curve is to be rotated

        Returns:
            Curve: A curve rotated theta radians relative to the origin
        """
        return Curve(
            self.P1.rotate(theta),
            self.P2.rotate(theta),
            self.P3.rotate(theta),
            self.P4.rotate(theta),
            self.speed,
            self.bp,
        )

    def scale(self, factor):
        """
        Returns a new Curve scaled by the specified factor.

        Args:
            factor (float): The factor with which to scale the Curve

        Returns:
            Curve: A curve scaled using the specified factor
        """
        return Curve(
            self.P1.scale(factor),
            self.P2.scale(factor),
            self.P3.scale(factor),
            self.P4.scale(factor),
            self.speed,
            self.bp,
        )

    def _bez(self, p0, p1, p2, p3, tc):
        t2 = tc * tc
        t3 = t2 * tc

        mt = 1 - tc
        mt2 = mt * mt
        mt3 = mt2 * mt

        return p0 * mt3 + 3 * p1 * mt2 * tc + 3 * p2 * mt * t2 + p3 * t3

    def arc_length_approximation(self):
        dist = 0
        curr_x = self.P1.get_x()
        curr_y = self.P1.get_y()
        approximation_steps = 32
        for i in range(approximation_steps):
            x = self._bez(
                self.P1.get_x(),
                self.P2.get_x(),
                self.P3.get_x(),
                self.P4.get_x(),
                i / approximation_steps,
            )
            y = self._bez(
                self.P1.get_y(),
                self.P2.get_y(),
                self.P3.get_y(),
                self.P4.get_y(),
                i / approximation_steps,
            )

            part_dist = math.sqrt(math.pow(curr_x - x, 2) + math.pow(curr_y - y, 2))

            dist = part_dist + dist
            curr_x = x
            curr_y = y

        return dist

    def get_segment_length(self) -> float:
        return self.arc_length_approximation()

    def get_segment_duration(self) -> float:
        return self.get_segment_length() / self.speed
